//
//  All designs and code in this project are distributed under the
//  Creative Commons Attribution-NonCommercial-ShareAlike 4.0
//  International license.
//
//  For more information visit http://creativecommons.org/licenses/by-nc-sa/4.0/
//
#if defined DISCOVERY
    #include <iostm8S105c6.h>
    #define UARTPORT(flag)    UART2_##flag
#else
    #include <iostm8s103f3.h>
    #define UARTPORT(flag)    UART1_##flag
#endif
#include <intrinsics.h>
#include <stdlib.h>
#include <string.h>

//
//  Define where we will be working in the EEPROM.
//
#define EEPROM_BASE_ADDRESS                 0x4000
#define EEPROM_PULSE_DURATION               ((unsigned char *) (EEPROM_BASE_ADDRESS + 0x10))
#define EEPROM_PRESCALAR_OFFSET             ((unsigned char *) (EEPROM_PULSE_DURATION + 2))
#define EEPROM_NUMBER_OF_COMMANDS_OFFSET    ((unsigned char *) (EEPROM_PRESCALAR_OFFSET + 2))
#define EEPROM_TIME_LAPSE_PAUSE_OFFSET      ((unsigned char *) (EEPROM_NUMBER_OF_COMMANDS + 1))
#define EEPROM_POWER_LED_OFFSET             ((unsigned char *) (EEPROM_TIME_LAPSE_PAUSE_OFFSET + 1))
#define EEPROM_SEQUENCE_LENGTH_TABLE_OFFSET ((unsigned char *) (EEPROM_BASE_ADDRESS + 0x20))
#define EEPROM_SEQUENCE_DATA_OFFSET         ((unsigned char *) (EEPROM_BASE_ADDRESS + 0x60))
#define EEPROM_LENGTH                       600
int _currentByte = 0;

//
//  Data ready for the pulse timer ISR's to use.
//
#define MAX_PULSE_NAME_LENGTH   8
int _numberOfPulses = 0;
int _currentPulse = 0;
unsigned char *_pulseDataAddress = NULL;

//
//  Some control variables to indicate if we are running or not.
//
#define STATE_WAITING_FOR_USER      0
#define STATE_RUNNING               1
int _currentState = STATE_WAITING_FOR_USER;

//
//  Working variables for the UART.
//
unsigned char *_currentTxByte;
unsigned short _currentTxCount;
unsigned short _txBufferLength;
#define UART_TX_BUFFER_SIZE         605
unsigned char _txBuffer[UART_TX_BUFFER_SIZE];
//
unsigned char *_currentRxByte;
unsigned short _currentRxCount;
unsigned short _rxBufferLength;
unsigned short _expectedRxDataLength = 0;
#define UART_RX_BUFFER_SIZE         605
unsigned char _rxBuffer[UART_RX_BUFFER_SIZE];
//
#define UART_MODE_WAITING_FOR_DATA          0
#define UART_MODE_RECEIVING_DATA            1
unsigned char _uartMode = UART_MODE_WAITING_FOR_DATA;

//
//  Commands which this remote control can understand.
//
#define COMMAND_GET_EEPROM_DATA         1
#define COMMAND_SET_EEPROM_DATA         2
#define COMMAND_TRANSMIT_PULSE_DATA     3
#define COMMAND_TRANSMIT_THIS_DATA      4
#define COMMAND_TIME_LAPSE              5
#define COMMAND_RESET                   6
#define COMMAND_POWER_LED_ENABLED       7

//
//  Error codes whch can be sent back to the calling application.
//
#define EC_OK                           0
#define EC_UNKNOWN_COMMAND              1
#define EC_RX_BUFFER_OVERFLOW           2
#define EC_CHECKSUM_FAILURE             3
#define EC_UNKNOWN_PULSE_SEQUENCE       4

//
//  Packet variables.
//
#define HEADER_LENGTH                   4
#define PACKET_OVERHEAD                 (HEADER_LENGTH + 1)
#define PACKET_HEADER                   0xaa

//--------------------------------------------------------------------------------
//
//  Calculate the checksum for the buffer.
//
unsigned char CalculateChecksum(unsigned char *buffer, short length)
{
    unsigned char checksum;

    checksum = 0xaa;
    for (int count = 0; count < length; count++)
    {
        checksum ^= *(buffer + count);
    }
    return(checksum);
}

//--------------------------------------------------------------------------------
//
//  Generic method for sending a response.
//
void SendResponse(unsigned char *buffer, unsigned short length)
{
    _txBufferLength = length;
    _currentTxByte = buffer;
    _currentTxCount = 0;
    UARTPORT(DR = PACKET_HEADER);
    UARTPORT(CR2_TIEN = 1);
}

//--------------------------------------------------------------------------------
//
//  Send an acknowledgement to the requester.
//
void SendACK(unsigned char command, unsigned char error)
{
    _txBuffer[0] = 0;
    _txBuffer[1] = 6;
    _txBuffer[2] = command;
    _txBuffer[3] = error;
    _txBuffer[4] = CalculateChecksum(_txBuffer, 4);
    SendResponse(_txBuffer, 5);
}

//--------------------------------------------------------------------------------
//
//  Write the a block of data into EEPROM.
//
void WriteDataToEEPROM(unsigned char *data, int length, unsigned char *destination)
{
    //
    //  Check if the EEPROM is write-protected.  If it is then unlock the EEPROM.
    //
    if (FLASH_IAPSR_DUL == 0)
    {
        FLASH_DUKR = 0xae;
        FLASH_DUKR = 0x56;
    }
    //
    //  Write the data to the EEPROM.
    //
    for (int index = 0; index < length; index++)
    {
        *destination++ = *data++;
    }
    //
    //  Now write protect the EEPROM.
    //
    FLASH_IAPSR_DUL = 0;
}

//--------------------------------------------------------------------------------
//
//  Setup the Rx buffers/counters ready to receive data.
//
void SetupRxBuffer()
{
    _currentRxByte = _rxBuffer;
    _currentRxCount = 0;
    _expectedRxDataLength = 0;
    _rxBufferLength = UART_RX_BUFFER_SIZE;
    _uartMode = UART_MODE_WAITING_FOR_DATA;
}

//--------------------------------------------------------------------------------
//
//  Transmit the IR pulse data.
//
int TransmitPulseData(unsigned char command)
{
    if (command < *((unsigned char *) EEPROM_NUMBER_OF_COMMANDS_OFFSET))
    {
        //
        //  We have enough command data.  Now work out where the command is.
        //
        _pulseDataAddress = (unsigned char *) (EEPROM_SEQUENCE_DATA_OFFSET);
        unsigned char *length = (unsigned char *) EEPROM_SEQUENCE_LENGTH_TABLE_OFFSET;
        for (unsigned char index = 0; index < command; index++)
        {
            _pulseDataAddress += *((unsigned char *) length);
        }
        //
        //  Now start processing the pulse data for the specified command.
        //
        _currentState = STATE_RUNNING;
        _pulseDataAddress += MAX_PULSE_NAME_LENGTH;     //  Skip the name.
        _numberOfPulses = *_pulseDataAddress++;
        _currentPulse = 0;
        TIM2_ARRH = *_pulseDataAddress++;
        TIM2_ARRL = *_pulseDataAddress++;
        //
        //  Now we have everything ready we need to force the Timer 2 counters to
        //  reload and enable Timers 1 & 2.
        //
        TIM2_CR1_URS = 1;
        TIM2_EGR_UG = 1;
        TIM1_BKR_MOE = 1;
//        TIM1_CR1_CEN = 1;
        TIM2_CR1_CEN = 1;
        return(1);
    }
    else
    {
        return(0);
    }
}

//--------------------------------------------------------------------------------
//
//  Process the data in the Rx buffer.
//
void ProcessUARTData()
{
    int dataLength = ((_rxBuffer[1] << 8) + _rxBuffer[2]) - 1;
    if (_rxBuffer[dataLength] == CalculateChecksum(_rxBuffer, dataLength))
    {
        switch (_rxBuffer[3])
        {
            case COMMAND_GET_EEPROM_DATA:
                _txBuffer[0] = ((EEPROM_LENGTH + PACKET_OVERHEAD) >> 8);
                _txBuffer[1] = ((EEPROM_LENGTH + PACKET_OVERHEAD) & 0xff);
                _txBuffer[2] = COMMAND_GET_EEPROM_DATA;
                memcpy(_txBuffer + 3, (const void *) EEPROM_BASE_ADDRESS, EEPROM_LENGTH);
                _txBuffer[UART_TX_BUFFER_SIZE - 1] = CalculateChecksum(_txBuffer, UART_TX_BUFFER_SIZE - 1);
                SendResponse(_txBuffer, UART_TX_BUFFER_SIZE);
                break;
            case COMMAND_SET_EEPROM_DATA:
                WriteDataToEEPROM(_rxBuffer + HEADER_LENGTH, dataLength, (unsigned char *) EEPROM_BASE_ADDRESS);
                SendACK(COMMAND_SET_EEPROM_DATA, EC_OK);
                break;
            case COMMAND_TRANSMIT_PULSE_DATA:
                if (TransmitPulseData(_rxBuffer[4]))
                {
                    SendACK(_rxBuffer[3], EC_OK);
                }
                else
                {
                    SendACK(_rxBuffer[3], EC_UNKNOWN_PULSE_SEQUENCE);
                }
                break;
            case COMMAND_TRANSMIT_THIS_DATA:
                break;
            case COMMAND_TIME_LAPSE:
                break;
            case COMMAND_RESET:
                break;
            case COMMAND_POWER_LED_ENABLED:
                break;
            default:
                SendACK(_rxBuffer[3], EC_UNKNOWN_COMMAND);
                break;
        }
    }
    else
    {
        SendACK(_rxBuffer[3], EC_CHECKSUM_FAILURE);
    }
    SetupRxBuffer();
}

//--------------------------------------------------------------------------------
//
//  Process the interrupt generated by the pressing of the button.
//
//  This ISR makes the assumption that we only have one incoming interrupt on
//  Port D.
//
//  Pressing the button on the remote control sends the first command sequence
//  held in the EEPROM.
//
#pragma vector = 8
__interrupt void EXTI_PORTD_IRQHandler(void)
{
    if (_currentState != STATE_RUNNING)
    {
        TransmitPulseData(0);
    }
}

//--------------------------------------------------------------------------------
//
//  Timer 2 Overflow handler.
//
#pragma vector = TIM2_OVR_UIF_vector
__interrupt void TIM2_UPD_OVF_IRQHandler(void)
{
    _currentPulse++;
    if (_currentPulse == _numberOfPulses)
    {
        //
        //  We have processed the pulse data so stop now.
        //
        PD_ODR_ODR3 = 0;
        TIM2_CR1_CEN = 0;
        TIM1_BKR_MOE = 0;
        _currentState = STATE_WAITING_FOR_USER;
    }
    else
    {
        TIM2_ARRH = *_pulseDataAddress++;
        TIM2_ARRL = *_pulseDataAddress++;
        TIM1_BKR_MOE = !TIM1_BKR_MOE;       //  Toggle the T1 output.
        TIM2_CR1_URS = 1;
        TIM2_EGR_UG = 1;
    }
    //
    //  Reset the interrupt otherwise it will fire again straight away.
    //
    TIM2_SR1_UIF = 0;
}

//--------------------------------------------------------------------------------
//
//  UART Transmit Buffer Empty handler.
//
#pragma vector = UARTPORT(T_TXE_vector)
__interrupt void UART_T_TXE_IRQHandler(void)
{
    if (_currentTxCount < _txBufferLength)
    {
        UARTPORT(DR = *_currentTxByte++);
        _currentTxCount++;
    }
    else
    {
        UARTPORT(CR2_TIEN = 0);
    }
}

//--------------------------------------------------------------------------------
//
//  UART Receive Buffer Not Empty handler.
//
#pragma vector = UARTPORT(R_RXNE_vector)
__interrupt void UART_R_RXNE_IRQHandler(void)
{
    unsigned char dataByte = UARTPORT(DR);
    if ((_uartMode == UART_MODE_WAITING_FOR_DATA) && (dataByte == PACKET_HEADER))
    {
        SetupRxBuffer();
        *_currentRxByte++ = dataByte;
        _currentRxCount++;
        _uartMode = UART_MODE_RECEIVING_DATA;
    }
    else
    {
        if (_uartMode == UART_MODE_RECEIVING_DATA)
        {
            if (_currentRxCount <= (UART_RX_BUFFER_SIZE - 1))
            {
                *_currentRxByte++ = dataByte;
                _currentRxCount++;
                if (_currentRxCount == 3)
                {
                    _expectedRxDataLength = (_rxBuffer[1] << 8) + _rxBuffer[2];
                }
                if (_currentRxCount > 3)
                {
                    if (_currentRxCount == _expectedRxDataLength)
                    {
                        ProcessUARTData();
                    }
                }
            }
            else
            {
                //
                //  If we get here we have filled the UART Rx buffer.
                //  Not a lot we can do really so reset the system to
                //  wait for a new command.
                //
                SendACK(_rxBuffer[3], EC_RX_BUFFER_OVERFLOW);
                _uartMode = UART_MODE_WAITING_FOR_DATA;
            }
        }
    }
}

//--------------------------------------------------------------------------------
//
//  Setup Timer 2 ready to process the pulse data.
//
void SetupTimer2()
{
    TIM2_PSCR = 4;
    TIM2_ARRH = 0;
    TIM2_ARRL = 0;
    TIM2_EGR_UG = 1;
    TIM2_IER_UIE = 1;       //  Enable the update interrupts.
}

//--------------------------------------------------------------------------------
//
//  Now set up the ports.
//
//  PD3 - IR Pulse signal.
//  PD4 - Input pin indicating that the user wishes to trigger the camera.
//
void SetupPorts()
{
    PD_ODR = 0;             //  All pins are turned off.
    //
    //  PD3 is the output for the IR control.
    //
    PD_DDR_DDR3 = 1;
    PD_CR1_C13 = 1;
    PD_CR2_C23 = 1;
    //
    //  Now configure the input pin (PD4).
    //
    PD_DDR_DDR4 = 0;        //  PD4 is input.
    PD_CR1_C14 = 1;         //  PD4 is floating input.
    PD_CR2_C24 = 1;         //  PD4 has external interrupt enabled.
    //
    //  Set up the interrupt on PD4.
    //
    EXTI_CR1_PDIS = 1;      //  Interrupt on rising edge.
    EXTI_CR2_TLIS = 1;      //  Rising edge only.
}

//--------------------------------------------------------------------------------
//
//  Set up Timer 1, channel 4 to output a PWM signal (the carrier signal).
//
void SetupTimer1()
{
    TIM1_ARRH = *((unsigned char *) EEPROM_PULSE_DURATION);
    TIM1_ARRL = *((unsigned char *) (EEPROM_PULSE_DURATION + 1));
    TIM1_PSCRH = *((unsigned char *) EEPROM_PRESCALAR_OFFSET);
    TIM1_PSCRL = *((unsigned char *) (EEPROM_PRESCALAR_OFFSET + 1));
    //
    //  Now configure Timer 1, channel 4.
    //
    TIM1_CCMR4_OC4M = 7;    //  Set up to use PWM mode 2.
    TIM1_CCER2_CC4E = 1;    //  Output is enabled.
    TIM1_CCER2_CC4P = 0;    //  Active is defined as high.
    //
    //  Work out the 50% duty cycle based upon the count.
    //
    unsigned short fiftyPercentDutyCycle;
    fiftyPercentDutyCycle = *((unsigned char *) EEPROM_PULSE_DURATION);
    fiftyPercentDutyCycle <<= 8;
    fiftyPercentDutyCycle += *((unsigned char *) (EEPROM_PULSE_DURATION + 1));
    fiftyPercentDutyCycle >>= 1;
    TIM1_CCR4H = ((fiftyPercentDutyCycle >> 8) & 0xff);
    TIM1_CCR4L = fiftyPercentDutyCycle & 0xff;
    //
    TIM1_BKR_MOE = 0;       //  Disable the main output.
    TIM1_CR1_CEN = 1;       //  Enable the timer.
}

//--------------------------------------------------------------------------------
//
//  Setup the UART to run at 115,200 baud, no parity, one stop bit, 8 data bits.
//
//  Important: This relies upon the system clock being set to run at 2 MHz.
//
void SetupUART()
{
    //
    //  Clear the Idle Line Detected bit in the status rerister by a read
    //  to the SR register followed by a Read to the DR register.
    //
    unsigned char tmp = UARTPORT(SR);
    tmp = UARTPORT(DR);
    //
    //  Reset the UART registers to the reset values.
    //
    UARTPORT(CR1 = 0);
    UARTPORT(CR2 = 0);
    UARTPORT(CR4 = 0);
    UARTPORT(CR3 = 0);
    UARTPORT(GTR = 0);
    UARTPORT(PSCR = 0);
    //
    //  Now setup the port to 115200,n,8,1.
    //
    UARTPORT(CR1_M = 0);        //  8 Data bits.
    UARTPORT(CR1_PCEN = 0);     //  Disable parity.
    UARTPORT(CR3_STOP = 0);     //  1 stop bit.
    UARTPORT(BRR2 = 0x0a);      //  Set the baud rate registers to 115200 baud
    UARTPORT(BRR1 = 0x08);      //  based upon a 16 MHz system clock.
    //
    //  Disable the transmitter and receiver.
    //
    UARTPORT(CR2_TEN = 0);      //  Disable transmit.
    UARTPORT(CR2_REN = 0);      //  Disable receive.
    //
    //  Set the clock polarity, lock phase and last bit clock pulse.
    //
    UARTPORT(CR3_CPOL = 1);
    UARTPORT(CR3_CPHA = 1);
    UARTPORT(CR3_LBCL = 1);
    //
    //  Turn on the UART transmit, receive and the UART clock.
    //
    UARTPORT(CR2_TEN = 1);
    UARTPORT(CR2_REN = 1);
    UARTPORT(CR3_CKEN = 0);
}

//--------------------------------------------------------------------------------
//
//  Setup the system clock to run at 16MHz using the internal oscillator.
//
void SetupSystemClock()
{
    CLK_ICKR = 0;                       //  Reset the Internal Clock Register.
    CLK_ICKR_HSIEN = 1;                 //  Enable the HSI.
    CLK_ECKR = 0;                       //  Disable the external clock.
    while (CLK_ICKR_HSIRDY == 0);       //  Wait for the HSI to be ready for use.
    CLK_CKDIVR = 0;                     //  Ensure the clocks are running at full speed.
    CLK_PCKENR1 = 0xff;                 //  Enable all peripheral clocks.
    CLK_PCKENR2 = 0xff;                 //  Ditto.
    CLK_CCOR = 0;                       //  Turn off CCO.
    CLK_HSITRIMR = 0;                   //  Turn off any HSIU trimming.
    CLK_SWIMCCR = 0;                    //  Set SWIM to run at clock / 2.
    CLK_SWR = 0xe1;                     //  Use HSI as the clock source.
    CLK_SWCR = 0;                       //  Reset the clock switch control register.
    CLK_SWCR_SWEN = 1;                  //  Enable switching.
    while (CLK_SWCR_SWBSY != 0);        //  Pause while the clock switch is busy.
}

//--------------------------------------------------------------------------------
//
//  Main program loop.
//
void main()
{
    __disable_interrupt();
    SetupSystemClock();
    SetupPorts();
    SetupUART();
    SetupTimer2();
    SetupTimer1();
    SetupRxBuffer();
    _uartMode = UART_MODE_WAITING_FOR_DATA;
    UARTPORT(CR2_RIEN = 1);
    __enable_interrupt();
    while (1)
    {
        __wait_for_interrupt();
    }
}